import zlib
import brotli
import gzip
from datetime import datetime
import json
from . import exceptions


class JsonException(exceptions.PrintableException):
    pass


class JsonEncodeException(JsonException):

    def __init__(self, obj, **kwargs):
        super().__init__(**kwargs)
        self._object = obj

    @property
    def object(self):
        return self._object


class JSONDecodeException(JsonException):

    def __init__(self, source, cause=None, **kwargs):
        self._source = source
        if cause is None:
            super().__init__()
        else:
            if not isinstance(cause, json.JSONDecodeError):
                raise TypeError("Cause must be json.JSONDecodeError")
            super().__init__(message=cause.msg, line=cause.lineno, column=cause.colno)

    @property
    def source(self):
        return self._source


strfyUnknownObjectTypes = False


class _JSONEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        if isinstance(obj, type):
            return obj.__name__
        try:
            if hasattr(obj, "_json"):
                return obj._json
        except Exception as e:
            raise JsonEncodeException(obj) from e
        try:
            return json.JSONEncoder.default(self, obj)
        except TypeError as e:
            if strfyUnknownObjectTypes:
                return str(obj)
            else:
                raise JsonEncodeException(obj) from e


def toJsonString(obj, indent=None, ensureAscii=True):
    return json.dumps(obj, cls=_JSONEncoder, indent=indent, ensure_ascii=ensureAscii)


def fromJsonString(js, allowEmpty=True):
    if allowEmpty and js == "" or js.isspace():
        return None
    try:
        return json.loads(js)
    except json.JSONDecodeError as e:
        raise JSONDecodeException(js, e)


class EncodingException(exceptions.PrintableException):
    pass


class StringEncodingException(EncodingException):
    pass


class UnknownCharsetException(StringEncodingException):

    def __init__(self, charset, **kwargs):
        if not isinstance(charset, str):
            raise TypeError("Charset must be str")
        super().__init__(charset=charset, message=f"Unknown charset '{charset}'", **kwargs)


def _stringEncoding(data, charset, decode):
    if not isinstance(charset, str):
        raise TypeError("Charset must be str")
    if decode:
        if not isinstance(data, bytes):
            raise TypeError("Data must be bytes")
    else:
        if not isinstance(data, str):
            raise TypeError("Data must be str")
    try:
        if decode:
            return data.decode(charset)
        else:
            return data.encode(charset)
    except LookupError:
        raise UnknownCharsetException(charset)
    except:
        raise StringEncodingException(charset=charset, message=f"Error while {'de' if decode else 'en'}coding string")


class DataEncodingException(EncodingException):
    pass


class UnknownEncodingTypeException(DataEncodingException):

    def __init__(self, encoding, **kwargs):
        if not isinstance(encoding, str):
            raise TypeError("Encoding must be str")
        super().__init__(encoding=encoding, message=f"Unknown encoding type '{encoding}'", **kwargs)


_dataEncoders = {
    "identity": lambda x: x,
    "gzip": gzip.compress,
    "deflate": zlib.compress,
    "br": brotli.compress
}

_dataDecoders = {
    "identity": lambda x: x,
    "gzip": gzip.decompress,
    "deflate": zlib.decompress,
    "br": brotli.decompress
}


def _dataEncoding(data, encoding, decode):
    if not isinstance(encoding, str):
        raise TypeError("Encoding type must be str")
    if not isinstance(data, bytes):
        raise TypeError("Data must be bytes")
    encoders = _dataDecoders if decode else _dataEncoders
    encoder = encoders.get(encoding.strip().lower(), None)
    if encoder is None:
        raise UnknownEncodingTypeException(encoding)
    try:
        return encoder(data)
    except:
        raise DataEncodingException(encoding=encoding, message=f"Error while {'de' if decode else 'en'}coding data'")


def encodeData(data, encoding):
    return _dataEncoding(data, encoding, False)


def encodeString(data, charset):
    return _stringEncoding(data, charset, False)


def decodeData(data, encoding):
    return _dataEncoding(data, encoding, True)


def decodeString(data, charset):
    return _stringEncoding(data, charset, True)


def encode(data, charset="utf-8", encoding="identity"):
    data = encodeString(data, charset)
    return encodeData(data, encoding)


def decode(data, charset="utf-8", encoding="identity"):
    data = decodeData(data, encoding)
    return decodeString(data, charset)


class NoCharsetSupportedException(StringEncodingException):
    pass


class NoEncodingSupportedException(DataEncodingException):
    pass


def tryEncode(data, suppCharsets=["utf-8"], suppEncodings=["identity"]):
    ok = False
    for charset in suppCharsets:
        try:
            data = encodeString(data, charset)
        except StringEncodingException:
            continue
        else:
            ok = True
            break
    if not ok:
        raise NoCharsetSupportedException()
    ok = False
    for encoding in suppEncodings:
        try:
            data = encodeData(data, encoding)
        except DataEncodingException:
            continue
        else:
            ok = True
            break
    if not ok:
        raise NoEncodingSupportedException()
    return (data, charset, encoding)



from . import exceptions
from . import encoding as swjasEncoding
import requests
from urllib import parse


class RequestException(exceptions.PrintableException):

    def __init__(self, requestBody, **kwargs):
        super().__init__(**kwargs)
        self._requestBody = requestBody

    @property
    def requestBody(self):
        return self._requestBody


class HttpException(RequestException):

    def __init__(self, requestBody, responseBody, statusCode, **kwargs):
        if not isinstance(statusCode, int):
            raise TypeError("Status code must be int")
        super().__init__(requestBody, statusCode=statusCode, statusMessage=exceptions.getStatusMessage(statusCode), **kwargs)
        self._responseBody = responseBody

    @property
    def responseBody(self):
        return self._responseBody


class TimeoutException(RequestException):

    def __init__(self, requestBody, timeout, **kwargs):
        super().__init__(requestBody, timeout=timeout, **kwargs)


class ConnectionException(RequestException):
    pass


class JSONDecodeException(RequestException):
    pass


class URLException(exceptions.PrintableException):

    def __init__(self, url, message, **kwargs):
        super().__init__(url=url, message=message, **kwargs)


def validateUrl(url):
    if not isinstance(url, str):
        raise TypeError("Expected str")
    url = parse.urlparse(url)
    if not url.netloc:
        raise URLException(url, "URL has no netloc")
    if any([url.query, url.params, url.fragment]):
        raise URLException(url, "URL cannot include parameters, query or fragment")
    if url.scheme:
        if url.scheme != "http":
            raise URLException(url, "Expected http scheme")
    else:
        url = url._replace(scheme="http")
    return url.geturl()


def validateHost(host):
    if not isinstance(host, str):
        raise TypeError("Expected str")
    host = parse.urlparse(host)
    if not host.netloc:
        raise URLException(host, "Host has no netloc")
    if any([host.path, host.query, host.params, host.fragment]):
        raise URLException(host, "Host cannot include path, parameters, query or fragment")
    if host.scheme and host.scheme != "http":
        raise URLException(host, "Expected http scheme")
    return host.netloc


def validateService(service):
    if not isinstance(service, str):
        raise TypeError("Expected str")
    service = parse.urlparse(service)
    if not service.path:
        raise URLException(service, "Service has no path")
    if any([service.netloc, service.scheme, service.query, service.params, service.fragment]):
        raise URLException(service, "Service cannot include scheme, netloc, parameters, query or fragment")
    return service.path


def service(host, service, **kwargs):
    # TODO Add docs

    host = validateHost(host)
    service = validateService(service)

    url = parse.urljoin(host.geturl(), service.geturl())

    return request(url, **kwargs)


def request(url, body=None, timeout=15, encoding="identity"):
    # TODO Add docs

    if not isinstance(url, str):
        raise TypeError("Url must be string")
    if not isinstance(timeout, (int, float)):
        raise TypeError("Timeout must be int or float")
    if timeout <= 0:
        raise ValueError("Timeout must be positive")

    url = validateUrl(url)

    body = swjasEncoding.toJsonString(body)
    body = swjasEncoding.encode(body, encoding=encoding)

    headers = {
        "Content-Type": "application/json;charset=utf-8",
        "Content-Encoding": encoding,
        "Accept-Charset": "utf-8",
        "Accept": "application/json"
    }

    try:
        res = requests.post(url, data=body, timeout=timeout, allow_redirects=False, headers=headers)
        if not res.ok:
            res.raise_for_status()
    except requests.HTTPError as e:
        try:
            resBody = swjasEncoding.fromJsonString(res.text)
        except swjasEncoding.JSONDecodeException:
            resBody = None
        raise HttpException(body, resBody, e.response.status_code) from None
    except requests.Timeout:
        raise TimeoutException(body, timeout) from None
    except requests.ConnectionError:
        raise ConnectionException(body) from None
    except requests.RequestException:
        raise RequestException(body) from None
    try:
        return swjasEncoding.fromJsonString(res.text, False)
    except swjasEncoding.JSONDecodeException as e:
        raise JSONDecodeException(body) from e


def _main():

    import argparse

    def makeArgparseRangeType(min, max, integer):

        def validate(arg):
            try:
                value = int(arg) if integer else float(arg)
            except ValueError:
                raise argparse.ArgumentTypeError(f"expected {'int' if integer else 'float'}")
            if not min <= value <= max:
                raise argparse.ArgumentTypeError(f"value not in range [{min},{max}]")
            return value

        return validate

    def argparseFileType(arg):
        import os
        path = os.path.abspath(arg)
        if not os.path.exists(path):
            raise argparse.ArgumentTypeError(f"file '{path}' does not exist")
        return arg

    def argparseUrlType(arg):
        try:
            return validateUrl(arg)
        except URLException as e:
            raise argparse.ArgumentTypeError(e.message)

    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("url", metavar="URL", help="Request URL", type=argparseUrlType)
    parser.add_argument("-if", "--inputfile", metavar="INPUT_FILE", default=None, help="Use an input file instead of stdin", type=argparseFileType)
    parser.add_argument("-to", "--timeout", default=15, metavar="TIMEOUT", help="Request timeout seconds", type=makeArgparseRangeType(1, 60, False))
    parser.add_argument("--indent", default=4, metavar="SPACES", help="Indentation tab size (-1 to minify)", type=makeArgparseRangeType(-1, 10, True))
    parser.add_argument("--outputstatus", action="store_true", help="Output a JSON object containing status info and response")
    parser.add_argument("--encoding", default="identity", metavar="ENCODING", choices=swjasEncoding._dataEncoders.keys(), help="Request encoding type")
    args = parser.parse_args()

    try:

        import sys
        import logging
        logger = logging.getLogger("swjas.client")

        if args.inputfile is not None:
            try:
                with open(args.inputfile, "r") as inputFile:
                    rawInput = inputFile.read()
            except IOError as e:
                logger.error("Error while reading from input file:\n%s", e.strerror)
                sys.exit(1)
            except UnicodeDecodeError as e:
                logger.error("Error while decoding input file text")
                sys.exit(1)
        else:
            rawInput = sys.stdin.read()

        jsonInput = swjasEncoding.fromJsonString(rawInput)

        def output(out):
            indent = args.indent if args.indent > 0 else None
            print(swjasEncoding.toJsonString(out, indent=indent))

        def outputResult(status, response):
            output({
                "statusCode": status,
                "statusMessage": exceptions.getStatusMessage(status),
                "data": response
            })
        
        def outputError(exception):
            output({
                "error": exception
            })

        try:
            responseBody = request(args.url, jsonInput, args.timeout, args.encoding)
        except HttpException as e:
            if args.outputstatus:
                outputResult(e.statusCode, e.responseBody)
            else:
                logger.error(e)
        except RequestException as e:
            if args.outputstatus:
                outputError(e)
            else:
                logger.error(e)
        else:
            outputResult(200, responseBody)

    except KeyboardInterrupt:
        logger.error("Interrupted by user")


if __name__ == "__main__":
    _main()



from urllib import parse as urlparse
from . import exceptions
from . import encoding
import logging
from datetime import datetime
import re

_logger = logging.getLogger("swjas")


def _cleanRoute(route):
    if (not isinstance(route, (tuple, list))) or len(route) != 2:
        raise TypeError("route must be a two-element tuple or list")
    path, handler = route
    if not isinstance(path, str):
        raise TypeError("route path must be a string")
    if not hasattr(handler, '__call__'):
        raise TypeError("route handler must be callable")
    url = urlparse.urlparse(path)
    if not url.path:
        raise ValueError("route path must be a valid network path")
    if any([url.scheme, url.netloc, url.fragment, url.query, url.params]):
        raise ValueError("route path cannot contain scheme, netloc or parameters infos")
    return (url.path.strip("/"), handler)


_headerWeightedListRegex = re.compile('(?P<value>[^;]*)(;q=(?P<weight>.*))?')
_mimeRegex = re.compile('(?P<type>[^;]*)[^(;charset=)]*(;charset=(?P<charset>.*))?')


def _parseMime(header):
    header = header.strip().lower()
    if header != "":
        match = _mimeRegex.match(header)
        if match is not None:
            t, c = match.group("type"), match.group("charset")
            if c is not None:
                if c.startswith("/"):
                    c = c[1:]
                if c.endswith("/"):
                    c = c[:-1]
            return (t, c)
    return (None, None)


def _parseHeaderList(header):
    return list(filter(None, map(str.strip, header.split(","))))


def _parseHeaderWeightedListItem(item):
    match = _headerWeightedListRegex.match(item.strip())
    if match is not None:
        v, w = match.group("value"), match.group("weight")
        if w is not None:
            try:
                w = float(w.strip())
            except ValueError:
                pass
            else:
                return (v, w)
        else:
            return (v, 1)
    return (None, None)


def _parseHeaderWeightedList(header, overrideWeight=None):
    if overrideWeight is None:
        overrideWeight = (lambda _, w: w)
    inList = _parseHeaderList(header)
    outList = []
    for item in inList:
        value, weight = _parseHeaderWeightedListItem(item)
        if value is not None:
            weight = overrideWeight(value, weight)
            outList += [(value, weight)]
    outList.sort(key=lambda i: -i[1])
    return list(map(lambda i: i[0], outList))


def makeApplication(routes, allowEmptyRequestBody=True):
    # TODO Add docs
    # Prepare route dict
    routeDict = {}
    for path, handler in map(_cleanRoute, routes):
        routeDict[path] = handler

    def application(environ, startResponse):

        # Capture path
        path = environ.get("PATH_INFO", "")
        if path.startswith("/") or path.startswith("\\"):
            path = path[1:]
        if path.endswith("/") or path.endswith("\\"):
            path = path[:-1]

        # Warn if JSON not accepted
        acceptedMimes = _parseHeaderList(environ.get("HTTP_ACCEPT", "").lower())
        if not any(mime in acceptedMimes for mime in ["*/*", "application/*", "application/json"]):
            _logger.info(f"Request to path '{path}' does not accept JSON reponse type: serving JSON anyway")

        # Find handler
        handler = routeDict.get(path)

        # Catch HTTP exceptions
        try:
            # Ensure POST method
            method = environ.get("REQUEST_METHOD")
            if method == "OPTIONS":
                responseBody = encoding.toJsonString(None)
                statusCode = 200
                statusMessage = exceptions.getStatusMessage(200)
            elif method == "POST":
                # Ensure no query
                query = environ.get("QUERY_STRING", "").strip()
                if query != "":
                    _logger.info(f"Rejected request to '{path}' with query '{query}'")
                    raise exceptions.BadRequestException(message="Unexpected query")

                if handler:
                    # Parse request JSON body
                    try:
                        requestBodyLength = int(environ.get('CONTENT_LENGTH', 0))
                        requestBody = environ['wsgi.input'].read(requestBodyLength)
                    except:
                        requestBody = ""

                    # Decode body
                    requestBodyType, requestBodyCharset = _parseMime(environ.get("CONTENT_TYPE", ""))

                    if requestBodyCharset is None:
                        _logger.info(f"Request to path '{path}' does not specify content charset: assuming utf-8")
                        requestBodyCharset = "utf-8"

                    if requestBodyType is None:
                        _logger.info(f"Request to path '{path}' does not specify content type: assuming JSON")
                    elif requestBodyType != "application/json":
                        _logger.info(f"Rejected request to '{path}' with non-JSON body type")
                        raise exceptions.HttpException.build(415, message="Expected JSON content type", requestContentType=requestBodyType)

                    requestBodyEncoding = environ.get("HTTP_CONTENT_ENCODING", "")
                    if requestBodyEncoding == "":
                        requestBodyEncoding = "identity"

                    try:
                        requestBody = encoding.decode(requestBody, requestBodyCharset, requestBodyEncoding)
                    except encoding.EncodingException as e:
                        _logger.info(f"Rejected request to '{path}' with undecodable body")
                        raise exceptions.BadRequestException(message="Unable to decode request body") from e

                    try:
                        jsonRequestBody = encoding.fromJsonString(requestBody, allowEmpty=allowEmptyRequestBody)
                    except encoding.JsonException as e:
                        _logger.info(f"Rejected request to '{path}' with invalid JSON body")
                        raise exceptions.BadRequestException(message="Error while decoding request JSON body") from e

                    # Call handler
                    try:
                        jsonResponseBody = handler(jsonRequestBody)
                        try:
                            responseBody = encoding.toJsonString(jsonResponseBody)
                        except:
                            _logger.error("Error while encoding JSON response body")
                            raise
                    except exceptions.HttpException as e:
                        raise e
                    except Exception as e:
                        _logger.exception(f"Exception while processing request to '{path}'")
                        raise exceptions.ServerErrorException(message="Error while processing the request")
                    else:
                        statusCode = 200
                        statusMessage = exceptions.getStatusMessage(200)

                else:
                    # No handler found
                    _logger.info(f"Rejected request to unrouted path '{path}'")
                    raise exceptions.NotFoundException(message="Invalid path", path=path)

            else:
                # Unsupported method
                _logger.info(f"Rejected request to '{path}' with method '{method}'")
                raise exceptions.HttpException.build(405, message="Expected POST method", requestedMethod=method)

        except exceptions.HttpException as e:
            # Prepare HTTP exception response
            def errorize(json):
                return {"error": json}

            try:
                statusCode = e.statusCode
                statusMessage = e.statusMessage
                responseBody = encoding.toJsonString(errorize(e))
            except encoding.JsonException:
                _logger.exception("Error while preparing JSON response for HttpException")
                fallbackException = exceptions.ServerErrorException(message="Error while collecting information about a previous error")
                statusCode = fallbackException.statusCode
                statusMessage = fallbackException.message
                responseBody = encoding.toJsonString(errorize(fallbackException))

        responseHeaders = []

        # Encode
        def overrideCharsetWeight(c, w): return 2 if c.lower() == "utf-8" else w
        acceptedCharsets = _parseHeaderWeightedList(environ.get("HTTP_ACCEPT_CHARSET", ""), overrideCharsetWeight)
        acceptedEncodingTypes = _parseHeaderWeightedList(environ.get("HTTP_ACCEPT_ENCODING", "")) + ["identity"]

        responseBody, chosenCharset, chosenEncoding = encoding.tryEncode(responseBody, acceptedCharsets + ["utf-8"], acceptedEncodingTypes)

        if not chosenCharset in acceptedCharsets:
            _logger.info(f"No accepted charset found for request to path '{path}': serving utf-8 anyway")

        responseHeaders += [("Content-Type", f"application/json;charset={chosenCharset}")]
        responseHeaders += [("Content-Encoding", chosenEncoding)]

        # Calculate content length
        responseBodyLength = len(responseBody) if responseBody is not None else 0
        responseHeaders += [("Content-Length", f"{responseBodyLength}")]

        # Provide allowed methods
        allow = "POST" if handler is not None else ""
        responseHeaders += [("Allow", allow)]
        if method == "OPTIONS":
            responseHeaders += [("Access-Control-Allow-Methods", allow)]
            responseHeaders += [("Access-Control-Allow-Headers", "Accept, Content-Encoding, Content-Type, Accept-Charset, Accept-Encoding")]

        # CORS
        responseHeaders += [("Access-Control-Allow-Origin", "*")]

        # Start response
        startResponse(f"{statusCode} {statusMessage}", responseHeaders)
        return [responseBody]

    return application
